unit baseimage;
{

  BMP reader/writer Based on Mazen Neifer's (LGPL) fpimage-fpwritebmp code in the FPC distribution.

}
interface
{$i vwrdefines.inc}
uses math,classes,types,sysutils, Windows,

     ctypes,candtype,bmpcomn;

Type
  tbw8image = Class
                 private
                  frefcount : Integer;  // first var because of alignment!
                  fstore : pbbyte;
                  allocptr:pbbyte;
                  fImageWidth : Integer;
                  fImageHeight : Integer;
                  fRowPitch : Integer;
                  falignment : integer;
                  baseptr,
                  endlineptr : pbbyte;
                  fTopDown : boolean;        // If set, bitmap is stored top down instead of bottom up
                  Rect : TRect;
                  lastalloc : integer;
                  procedure settopdown(const Value: boolean);
                  function getreservedsize: integer;
                 protected
                  BFI : TBitMapInfoHeader;  // The header as read from the stream.
                  StartPosition : int64; { save start of bitmap in the stream, if we must go back and fix something }
                  FBpp : byte;

                  BFH : TBitMapFileHeader;
                  RedMask, GreenMask, BlueMask : longword; //Used if Compression=bi_bitfields
                  RedShift, GreenShift, BlueShift : shortint;

                  // SetupRead will allocate the needed buffers, and read the colormap if needed.
                  procedure SetupRead(nPalette, nRowBits: Integer; Stream : TStream); virtual;
                  function CountBits(Value : byte) : shortint;
                  function ShiftCount(Mask : longword) : shortint;
                 // procedure ExpandRLE8ScanLine(Row : Integer; Stream : TStream);
                 // required by TFPCustomImageReader
                  function  InternalCheck (Stream:TStream) : boolean;
                  function  SaveHeader(Stream:TStream):boolean;
                  procedure SetImageWidth( fww: Integer );
                  procedure SetImageHeight( fwh: Integer );
                  procedure alloc;
                 public
                  machinespeed : integer;
                  timestamp : int64;
                  //machinespeed : int64;
                  {$ifndef userefcount}
                  inpool:boolean;
                  {$endif}
                  function getbaseptr : pbbyte;
                  destructor destroy; override;
                  constructor create(p:pointer); // allow NIL, but nothing else
                  procedure kopieerlines(source: tbw8image;van,totenmet:integer;naar:integer=0);
                  procedure clear; overload;
                  procedure clear(vanline,totenmet:integer); overload;
                  procedure kopieer(source: tbw8image); overload;
                  procedure setsize(width,height:integer);
                  procedure LoadFromStream(Stream:TStream);
                  procedure SaveToStream(Stream:TStream);
                  procedure savefile(s:string);
                  procedure loadfile(s:string);
                  procedure increfcount;
                  function  decrefcount:integer;
                  procedure clearrefcount(val:integer=1);
                  function GetImagePointer( OriginX, OriginY : Integer ): pbbyte; {$ifndef ver150} inline;{$endif}
                  function GetLinePointer(originy:integer):pbbyte; {$ifndef ver150} inline;{$endif}
                  property scanline[i:integer]:pbbyte read getlinepointer;
                  property RowPitch : Integer read fRowPitch;
                  property ImageHeight : Integer read fImageHeight write SetImageHeight;
                  property ImageWidth : Integer read fImageWidth write SetImageWidth;
                  property Height : Integer read fImageHeight write SetImageHeight;
                  property Width : Integer read fImageWidth write SetImageWidth;
                  property topdown : boolean read fTopDown write settopdown;
                  property raw : pbbyte read getbaseptr;
                  property reservedsize : integer read getreservedsize;
                  property refcount : Integer read frefcount;
                 end;

implementation
uses candelautils {$ifdef newmove} ,movejohunit10 {$endif};

Const
  aligndefault = 64;
  alignpower2minus1 = aligndefault-1; // and mask

{ tbw8image }

procedure tbw8image.alloc;
var i,alignbytes : integer;

begin
  if ftopdown then i:=1 else i:=-1;
  fRowPitch:=i*((fimagewidth+falignment-1) and not (falignment-1) );     // rowpitch = width rounded up for alignemnt times existing sign.
  i:=abs(fRowPitch)*fimageheight;
  if i<>lastalloc then
    begin
      try
      assert(i<30000000,'really big'+inttostr(i));
      ReallocMem(allocptr,i+aligndefault);
      alignbytes:=aligndefault-(ptruint(allocptr) and alignpower2minus1);
      fstore:=@allocptr[alignbytes];
      lastalloc:=i;
      except
      on e: exception do e.message:='Error reallocating size='+inttostr(i)+' rowpitch='+inttostr(frowpitch)+' imageheight='+inttostr(fimageheight);
      end;
    end;
  endlineptr:=nil;
  baseptr:=nil;
  if (fImageHeight>0) and (fImageWidth>0) then
   begin

     endlineptr:=@fstore[(imageheight-1)*abs(frowpitch)];
     if ftopdown then
       baseptr:=@fstore[0]
     else
       baseptr:=endlineptr
   end;
end;

constructor tbw8image.create;
begin
  falignment:=4;
  fImageWidth:=0;
  fimageheight:=0;
  frowpitch:=0;
  endlineptr:=nil;
  baseptr:=nil;
  lastalloc:=0;
  frefcount:=0;
end;

function tbw8image.decrefcount: integer;
begin
 result:=InterlockedDecrement(frefcount);
end;

destructor tbw8image.destroy;
begin
  freemem(allocptr);
  inherited;
end;

function tbw8image.getbaseptr: pbbyte;
begin
 result:=@fstore[0];
end;

function tbw8image.GetImagePointer(OriginX, OriginY: Integer): pbbyte;
begin
  result:=nil;
  if assigned(baseptr) then
     Result := pbbyte(@baseptr[ frowpitch*originy+originx])
end;

function tbw8image.GetLinePointer(originy: integer): pbbyte;
begin
  result:=nil;
  if assigned(baseptr) then
     Result := pbbyte(@baseptr[ frowpitch*originy])
end;

function tbw8image.getreservedsize: integer;
var i : integer;
begin
//  if ftopdown then i:=1 else i:=-1;
//  fRowPitch:=i*((fimagewidth+falignment-1) and not (falignment-1) );     // rowpitch = width rounded up for alignemnt times existing sign.
  result:=abs(fRowPitch)*fimageheight;
end;

procedure tbw8image.increfcount;
begin
 InterlockedIncrement(frefcount);
end;

procedure tbw8image.SetImageHeight(fwh: Integer);
begin
  fImageHeight:=fwh;
  alloc;
end;

procedure tbw8image.SetImageWidth(fww: Integer);
begin
  fimagewidth:=fww;
  alloc;
end;

procedure tbw8image.setsize(width, height: integer);
begin
  if (fImageWidth<>width) or (fImageHeight<>height) then
    begin
      fimagewidth:=Width;
      fimageheight:=height;
      alloc;
    end;
end;

procedure tbw8image.settopdown(const Value: boolean);
begin
  fTopDown := Value;
  alloc;
end;

var Colinfo,Colinfodummy : array[0..255] of TColorRGBA;

function tbw8image.SaveHeader(Stream:TStream):boolean;
begin
  FBpp :=8; // hardcoded for now.
  Result:=False;
  with BFI do
    begin
     Size:=sizeof(TBitMapInfoHeader);
     Width:=imageWidth;
     if not ftopdown then
       Height:=imageHeight    // ?
     else
       Height:=-imageHeight;
     Planes:=1;
     if FBpp=15 then BitCount:=16
     else BitCount:=FBpp;
     XPelsPerMeter:=100;
     YPelsPerMeter:=100;
     ClrImportant:=0;
    end;
  with BFH do
    begin
     bfType:=BMmagic;//'BM'
     bfOffset:=sizeof(TBitMapFileHeader)+sizeof(TBitMapInfoHeader)+length(ColInfo)*4;
     bfReserved:=0;
     bfSize:=bfOffset+BFI.SizeImage;
    end;
   StartPosition:=Stream.Position;
   Stream.Write(bfh,sizeof(TBitMapFileHeader));
   Stream.Write(bfi,sizeof(TBitMapInfoHeader));
   Result:=true;
end;


procedure tbw8image.savefile(s: string);
var
  Stream: TStream;
begin
  Stream:=TFileStream.Create(s, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure  tbw8image.SaveToStream(Stream:TStream);
var
  Row,Col,RowSize:Integer;
  PadCount : byte;
  aLine: PByte;
  i : Integer;
  tmppos : int64;
  Rect : TRect;
begin
    Rect.Left:=0; Rect.Top:=0; Rect.Right:=0; Rect.Bottom:=0;
    BFI.Compression:=BI_RGB;
    BFI.ClrUsed:=0;
    RowSize:=imageWidth;
    PadCount:=(4-(RowSize mod 4)) mod 4;   // rowpitch
    inc(RowSize,PadCount);
    BFI.SizeImage:=RowSize*imageHeight;
    SaveHeader(Stream); { write the headers }
    stream.Write(colinfo[0],sizeof(TColorRGBA)*length(colinfo));
    stream.Write(fstore[0],bfi.sizeimage);
end;

{ Counts how many bits are set }
procedure tbw8image.clear(vanline, totenmet: integer);
var
  RowPitchSource: Integer;
  pSource : pbbyte;
  top : LongWord;
begin
  rowpitchsource:=rowpitch;
  psource:= pbbyte(getImagePointer(0,vanline));
  if (imagewidth=0) or (imageheight=0) then exit;

  top:=totenmet-vanline+1;
  while top>0 do
    begin
      fillchar(psource[0],imagewidth,#128);
      inc(pSource,RowPitchSource);
      dec(top);
    end;
end;

procedure tbw8image.clearrefcount(val: integer);
begin
 frefcount:=val;
end;

function tbw8image.CountBits(Value : byte) : shortint;
var i,bits : shortint;
begin
  bits:=0;
  for i:=0 to 7 do
  begin
    if (value mod 2)<>0 then inc(bits);
    value:=value shr 1;
  end;
  Result:=bits;
end;

{ If compression is bi_bitfields, there could be arbitrary masks for colors.
  Although this is not compatible with windows9x it's better to know how to read these bitmaps
  We must determine how to switch the value once masked
  Example: 0000 0111 1110 0000, if we shr 5 we have 00XX XXXX for the color, but these bits must be the
  highest in the color, so we must shr (5-(8-6))=3, and we have XXXX XX00.
  A negative value means "shift left"  }
function tbw8image.ShiftCount(Mask : longword) : shortint;
var tmp : shortint;
begin
  tmp:=0;
  if Mask=0 then
  begin
    Result:=0;
    exit;
  end;

  while (Mask mod 2)=0 do { rightmost bit is 0 }
  begin
    inc(tmp);
    Mask:= Mask shr 1;
  end;
  tmp:=tmp-(8-CountBits(Mask and $FF));
  Result:=tmp;
end;

procedure tbw8image.SetupRead(nPalette, nRowBits: Integer; Stream : TStream);

var
  ColInfo: ARRAY OF TColorRGBA;
  i: Integer;

begin
  if ((BFI.Compression=BI_RGB) and (BFI.BitCount=16)) then { 5 bits per channel, fixed mask }
  begin
    RedMask:=$7C00; RedShift:=7;
    GreenMask:=$03E0; GreenShift:=2;
    BlueMask:=$001F; BlueShift:=-3;
  end
  else if ((BFI.Compression=BI_BITFIELDS) and (BFI.BitCount in [16,32])) then { arbitrary mask }
  begin
    Stream.Read(RedMask,4);
    Stream.Read(GreenMask,4);
    Stream.Read(BlueMask,4);
    RedShift:=ShiftCount(RedMask);
    GreenShift:=ShiftCount(GreenMask);
    BlueShift:=ShiftCount(BlueMask);
  end
  else if nPalette>0 then
    begin
    if BFI.ClrUsed>0 then
      Stream.Read(ColInfodummy[0],BFI.ClrUsed*SizeOf(TColorRGBA))
    else // Seems to me that this is dangerous.
      Stream.Read(ColInfodummy[0],nPalette*SizeOf(TColorRGBA));
    end
  else if BFI.ClrUsed>0 then { Skip palette }
    Stream.Position := Stream.Position + BFI.ClrUsed*SizeOf(TColorRGBA);
 if ftopdown then i:=1 else i:=-1;
 frowpitch:=i*((nRowBits + 31) div 32) shl 2;
end;

procedure tbw8image.LoadFromStream(Stream:TStream);

Var
  Row, i, pallen : Integer;
  BadCompression : boolean;
  p:pbbyte;
begin
  internalcheck(stream);
  Rect.Left:=0; Rect.Top:=0; Rect.Right:=0; Rect.Bottom:=0;
  Stream.Read(BFI,SizeOf(BFI));
  Stream.Position:=Stream.Position-SizeOf(BFI)+BFI.Size;
  with BFI do
  begin
    BadCompression:=false;
    if ((Compression=BI_RLE4) and (BitCount<>4)) then BadCompression:=true;
    if ((Compression=BI_RLE8) and (BitCount<>8)) then BadCompression:=true;
    if ((Compression=BI_BITFIELDS) and (not (BitCount in [16,32]))) then BadCompression:=true;
    if not (Compression in [BI_RGB..BI_BITFIELDS]) then BadCompression:=true;
    if BadCompression then
      raise Exception.Create('Bad BMP compression mode');
    fTopDown:=(Height<0);
    Height:=abs(Height);
    if (fTopDown and (not (Compression in [BI_RGB,BI_BITFIELDS]))) then
      raise Exception.Create('Top-down bitmaps cannot be compressed');
//    SetSize(0,0);

    {8:}
     SetupRead(256,Width*8,Stream);
  end;

{    pallen:=0;
    if BFI.BitCount<=8 then
      if BFI.ClrUsed>0 then pallen:=BFI.ClrUsed
      else pallen:=(1 shl BFI.BitCount);
}      
    SetSize(BFI.Width,BFI.Height);

    stream.Read(fstore[0],abs(frowpitch)*imageheight)
end;

function  tbw8image.InternalCheck (Stream:TStream) : boolean;
begin
  stream.Read(BFH,SizeOf(BFH));
  With BFH do
    Result:=(bfType=BMmagic); // Just check magic number
end;

procedure tbw8image.kopieer(source: tbw8image);
var
  RowPitchSource, RowPitchTarget : Integer;
  pSource, pTarget : pbbyte;
  i : LongWord;
begin
  timestamp:=source.timestamp;
  setsize(source.imagewidth,source.imageheight);
  RowPitchSource := source.rowpitch;
  RowPitchTarget := self.RowPitch;
  pSource := source.GetLinePointer(0);
  pTarget := GetLinePointer(0);
  if (imagewidth=0) or (imageheight=0) then exit;
  for i := 0 to ImageHeight-1 do
    begin
      {$ifdef newmove}Move_JOH_SSE_10{$else}move{$endif}(pSource[0],pTarget[0],imagewidth);
      inc(pSource,RowPitchSource);
      inc(pTarget,RowPitchTarget);
    end;
end;


procedure tbw8image.kopieerlines(source: tbw8image;van,totenmet:integer;naar:integer=0);
var
  RowPitchSource, RowPitchTarget : Integer;
  pSource, pTarget : pbbyte;
  i : LongWord;
begin
  if imagewidth<>source.imagewidth then
    imagewidth:=source.imagewidth;
  RowPitchSource := source.rowpitch;
  RowPitchTarget := self.RowPitch;
  pSource := pbbyte(source.getImagePointer(0,van));
  pTarget := pbbyte(getImagePointer(0,naar));
  if (imagewidth=0) or (imageheight=0) then exit;
  for i := van to totenmet do
    begin
      {$ifdef newmove} Move_JOH_SSE_10{$else} move{$endif}(pSource[0],pTarget[0],imagewidth);
      inc(pSource,RowPitchSource);
      inc(pTarget,RowPitchTarget);
    end;
end;

procedure tbw8image.loadfile(s: string);
var
  Stream: TStream;
begin
  Stream:=TFileStream.Create(s, fmOpenRead);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure inittable;
var i : integer;
begin
  for i := 0 to 255 do
    with colinfo[i] do
      begin
        B:=i;
        G:=i;
        R:=i;
        A:=0;
      end;
end;

procedure tbw8image.clear;
var
  i : integer;
begin
  i:=lastalloc;
  if i<>0 Then
    fillchar(fstore[0],i,#128);
end;

initialization
 inittable;
end.
