unit basestreaming;

interface
{$i vwrdefines.inc}
Uses
  Classes,contnrs;

Type
     TRoiRect     = record
                      ROIOrgX, ROIOrgY,ROIWidth, ROIHeight: Integer;  
                    end;


     TMyReader     = class; // vooruit definieren deze klassen, want deze hebben Tbasisobject nodig.
     TMyWriter     = class; // maar tbasisobject heeft ook deze nodig.


     TBasisStreamableobject= class
                         procedure schrijf(t:TMyWriter); virtual;
                         procedure laad (t:TmyReader); virtual;
                         constructor create; virtual;
                         destructor destroy; override;
                      end;
     TBasisStreamableobjectClass = Class of TBasisStreamableobject;
     TVersionedObject = class // (TBasisStreamableObject?)
                           mversie : Integer;
                           function ResolveClass(typ:ansistring):TBasisStreamableobjectClass; virtual; abstract;
                           function ResolveValue(typ:integer):TBasisStreamableobjectClass; virtual; abstract;
                         end;

     TMyReader = class
                   private
                     procedure fixup(var buffer;asize,realbytes:integer);
                   public
                     handle : TStream;
                     fhier   : TVersionedObject;
                     procedure open(fn:ansistring);
                     procedure openmem(mem:TStream);
                     procedure close;
                     procedure readstring(var s:ansistring);
                     procedure readinteger(var i:integer);
                     procedure readsingle(var s:single);
                     procedure readdouble(var s:double);
                     procedure readboolean(var b : boolean);
                     procedure readvariant(var v:variant);
                     procedure readbin    (var buffer;asize:Integer);
                     procedure readsizedbin    (var buffer;asize:Integer);
                     procedure readstringlist (var sl :TStringList);
                     function  readarraybasisobject:TObjectList;
                     procedure setversionobject(hier:TVersionedObject);
                     procedure rewind;   // alleen voor filebased;
                  end;

     TMyWriter = class
                     handle : TStream;
                     procedure open(fn:ansistring);
                     procedure openmem(strm:TStream);
                     procedure close;
                     procedure writestring(s:ansistring);
                     procedure writeinteger(i:integer);
                     procedure writesingle(s:single);
                     procedure writedouble(s:double);
                     procedure writeboolean(b : boolean);
                     procedure writevariant(v:variant);
                     procedure writebin(const buffer;asize:Integer);
                     procedure writesizedbin(const buffer;asize:Integer);
                     procedure writestringlist(sl:tstringlist);
                     procedure writearraybasisobject(lst:TObjectList);
                  end;

implementation

Uses variants,sysutils;

procedure TMyReader.close;
begin
  handle.free;
end;

procedure TMyReader.open(fn: ansistring);
begin
  handle:=TfileStream.Create(fn,fmopenread);
end;

procedure TMyReader.openmem(mem: TStream);
begin
  handle:=mem;
  handle.Position:=0;
end;

function TMyReader.readarraybasisobject:TObjectList;
var
  b   : TBasisStreamableobject;
  i,j,
  cnt : integer;
  typ : integer;
  s   : ansistring;
  thetype : TBasisStreamableobjectClass;
begin
  result:=tobjectlist.Create;
  result.OwnsObjects:=true;
  readinteger(cnt);
  if cnt >0 then
  for i:=0 to cnt-1 do
    begin
      readstring(s);
      thetype:=fhier.ResolveClass(s);
      if not assigned(thetype) then
        raise Exception.create('onbekende meting: '+s);
      b:=thetype.Create;
      b.laad(self);
      result.add(b);
    end;
end;

procedure TMyReader.readboolean(var b: boolean);
var i : integer;
begin
  readinteger(i);
  b:=i=1;
end;

procedure TMyReader.readdouble(var s: double);
begin
  handle.Read(s,sizeof(double));      // sizeof(integer)= MEESTAL 4
end;

procedure TMyReader.readinteger(var i: integer);
begin
  handle.Read(i,sizeof(integer));      // sizeof(integer)= MEESTAL 4
end;

procedure TMyReader.readsingle(var s: single);
begin
  handle.Read(s,sizeof(single));      // sizeof(integer)= MEESTAL 4
end;

procedure TMyReader.readstring(var s: ansistring);
var
  len : integer;
begin
  readinteger(len);
  setlength(s,len);
  if len<> 0 then
    handle.read(s[1],len);
end;

procedure TMyReader.readvariant(var v: variant);
var
  t : tvartype;
  d : double;
  i : integer;
  s : ansistring;
begin
  readinteger(i);
  t:=tvartype(i);
  case t of
    varDouble : begin
                  readdouble(d);
                  v:=d;
                end;
    varInteger: begin
                  readinteger(i);
                  v:=i;
                end;
    varString : begin
                  readstring(s);
                  v:=s;
                end;
    varboolean : begin
                   readinteger(i);
                   v:=i=1;
                 end;
  end;
end;

procedure TMyReader.readbin(var buffer; asize: Integer);
begin
  handle.Read(buffer,asize)
end;

procedure TMyReader.rewind;
begin
 handle.Seek(0,soBeginning);
end;

procedure TMyReader.setversionobject(hier: TVersionedObject);
begin
  fhier:=hier;
end;



{ TMyWriter }
procedure TMyWriter.close;
begin
  freeandnil(handle);
end;

procedure TMyWriter.open(fn: ansistring);
begin
  handle:=TfileStream.Create(fn,fmcreate or fmShareDenyNone);
end;

procedure TMyWriter.writeboolean(b: boolean);

var i : integer;
begin
 if b then i:=1 else i:=0;
 writeinteger(i);
end;

procedure TMyWriter.writearraybasisobject(lst: TObjectList);
var
  i :integer;
begin
  writeinteger(lst.Count);
  if lst.count> 0 then
    begin
      for i:=0 to lst.count-1 do
        begin
//          if meetmodelversie>$160010 then
            self.writestring(TBasisStreamableobject(lst[i]).ClassName);
  {        else
             writeinteger(tbasisobject(lst[i]).objecttypeidentifier);}
          TBasisStreamableobject(lst[i]).schrijf(self);
        end;
    end;
end;

procedure TMyWriter.writeinteger(i: integer);
begin
  handle.write(i,sizeof(i));
end;

procedure TMyWriter.writesingle(s: single);
begin
  handle.write(s,sizeof(single));
end;

procedure TMyWriter.writestring(s: ansistring);
var
  len:integer;
begin
  len:=length(s);
  handle.write(len,sizeof(integer));
  if len>0 then
    handle.write(s[1],len);
end;

procedure TMyWriter.writedouble(s: double);
begin
  handle.write(s,sizeof(double));
end;

procedure TMyWriter.openmem(strm:TStream);
begin
  handle:=strm;
end;

procedure TMyWriter.writebin(const buffer;asize:integer);
begin
  handle.write(buffer,asize);
end;

procedure TMyWriter.writevariant(v: variant);
var
  t:tvartype;
begin
  t:=vartype(v);
  writeinteger(ord(t));
  case t of
    varDouble : writedouble(v);
    varInteger: writeinteger(v);
    varstring : writestring(v);
    varboolean: if v then
                  writeinteger(1)
                 else
                   writeinteger(0);
  end;
end;

procedure TMyWriter.writestringlist(sl: tstringlist);
var i,cnt : integer;
    s : ansistring;
begin
 cnt:=sl.count;
 writeinteger(cnt);
 if cnt>0 then
   for i:=0 to cnt-1 do
     writestring(sl[i]);
end;

procedure TMyWriter.writesizedbin(const buffer; asize: Integer);
// this function is mainly for stuff like sets that can suddenly grow when members are added.
var b: byte;
begin
  b:=asize;           // endianess
  handle.Write(b,1);
  handle.write(buffer,asize);
end;

{ TBasisStreamableobject }

constructor TBasisStreamableobject.create;
begin

end;

destructor TBasisStreamableobject.destroy;
begin

end;

procedure TBasisStreamableobject.laad(t: TmyReader);
begin

end;

procedure TBasisStreamableobject.schrijf(t: TMyWriter);
begin

end;

procedure TMyReader.readstringlist(var sl: TStringList);
var i,cnt : integer;
    s : ansistring;
begin
 sl.clear;
 readinteger(cnt);
 if cnt>0 then
   for i:=0 to cnt-1 do
     begin
       readstring(s);
       sl.Add(s);
     end;
end;

procedure TMyReader.readsizedbin(var buffer; asize: Integer);
// we assume little endian set packing. If not, fix to read the bytes at buffer[asize-b] or so.
var b: byte;

begin
  fillchar(buffer,asize,#0);
  handle.read(b,1);
  if b<=asize then
    handle.Read(buffer,b)
  else
    fixup(buffer,asize,b);
end;

procedure TMyReader.fixup(var buffer; asize, realbytes: integer);
// we assume here little endian set packing.
var
   toread:integer;
   dummybuffer : array[0..255] of byte;
begin
  // asize<b
  handle.read(buffer,asize);
  dec(realbytes,asize);
  while realbytes>0 do
    begin
      toread:=realbytes;
      if toread>sizeof(dummybuffer) then
        toread:=sizeof(dummybuffer);
      handle.read(dummybuffer,toread);
      dec(realbytes,toread);
    end;
end;

end.
