UNIT glBlitter;

{$define debug} // remove to get rid of timers

interface
{$i vwrdefines.inc}
uses
  SysUtils, Classes, Windows, Controls, StdCtrls, ExtCtrls, Forms, Graphics, Dialogs,
  dglOpenGL,baseoverlay,candelautils;

Type
  tfontsizecache = array[0..nrOverlayLayers] of integer;
  TBlitTimings = (blitrefresh1,blitrefresh2,blitrefresh3,blitrefresh4,blitrefreshhud,blittexture);
  TGLBlitter = class;
  TGLOverlay = class
   protected
    ovldata:TOverlayLayerData;
    FOwner: TGLBlitter;
//    FListStatus: TGLOverlayStatus;
    FVisible: Boolean;
    FLineWidth: GLfloat;
    FFontBase: GLuint;

    FFont: TFont;
    FRectVBO: GLuint; // quad vertices
    FRectCBO: GLuint; // quad colors

    FLineVBO: GLuint; // 1px line vertices
    FLineCBO: GLuint; // 1px line colors

    FThickLineVBO: GLuint; // Xpx line vertices
    FThickLineCBO: GLuint; // Xpx line colors

    FStippleLineVBO: GLuint; // Xpx line vertices
    FStippleLineCBO: GLuint; // Xpx line colors

    FTriangleVBO: GLuint; // filled ellipses
    FTriangleCBO: GLuint; // filled ellipses colors
    fattached : boolean;
    { Creates new images for font and adds to display list for GL to draw }
    procedure BuildFont;
    { These 2 are for text drawing, since text is drawn in immediate mode
      I decided to put it here, rather than in specific overlays }
    procedure glPrint(const aText: ansistring);
    procedure TextOut(const i: Integer);
    { Common ancestor method for clearing the flush on 1st call
      to DrawPrimitive() }
    { Sets font in a very thread unfriendly way :) }
    procedure SetFont(aFont:TFont);
    { Frees the font GL stuff }
    procedure KillFont;
    { Makes sure it's refreshed }
    procedure SetVisible(const aValue: Boolean);
   public
    { Create the overlay over an existing blitter }
    constructor Create(aOwner: TGLBlitter);
    destructor Destroy; override;
    procedure assign(gl:TOverlaylayerdata);
    { Base method for flushing given data. This should be called when all
      data is set for a draw }
    procedure Flush;
    { Base method for actual drawing of the overlay. Works only if flush was
      called properly }
    procedure DrawOverlay;
    procedure clear;
    procedure colorselection(col:tcolor);

    procedure attach;
    procedure detach;
    { Can be turned off so you can specify which overlays to show }
    property Visible: Boolean read FVisible write SetVisible;
    { Global linewidth for all ThickLines of this overlay }
    property LineWidth: GLfloat read FLineWidth write FLineWidth;
    { Global font for all texts of this overlay }
    property Font: TFont read FFont write SetFont;
  end;



  { This is THE VBO overlay, the one you should use in the
    blitter. It uses Vertex Buffer Objects for primitive
    data, so all the stuff you want to draw is flushed
    in just 1 GL call, and also drawn in a fast way.

    Two things to note for future fiddling:
    1. Always use GLfloat for vertex data
    2. Always use 4 bytes for color data (3 bytes -> bad performance,
                                         even if you don't use alpha)
    #1 is because todays cards don't implement double precision in HW,
    #2 is because they need aligned data of 4byte sizes, and 3 singles per
       color would be simply too much to waste, so 4 bytes are perfect. }

  { This is the overlay class, currently "on top" of the panel.
    If you need Pain() you'll need to find a way to either .Assign
    or construct the thing in a different style. It's already
    inherited from TPanel so it shouldn't be much work. I also
    made the main "refresh" to be Paint, and refresh just calls it.
    All that's required is to remove the FPanel stuff (about 3 uses, width/height stuff)
    and create it in a manual way runtime.

    Not much has changed since last time, only added the overlay
    array (count defined in constant on top of unit) }
  TGLBlitter = class(TPanel)
   protected
    FYOffset: GLfloat;
    FXOffset: GLfloat;
    FLightAmbient: array[0..3] of GLfloat;
    FPictureWidth, FPictureHeight: Integer;
    FTexture: GLuint;
    FRC: HGLRC;
    FDC: HDC;
    FBrightness: GLfloat;
    FBrightnessUniform: GLint;
    FZoomX: GLfloat;
    FZoomY: GLfloat;
    FMaxTextureSize: GLint;
    FOverlay: array[1..NrOverlayLayers] of TGLOverlay;
    initdone:boolean;
    faspectX,
    faspectY : single;
    foverrideaspect:boolean;
    textureallocated : boolean;
    FHud: TGLOverlay;
    picloaded : boolean;
    procedure SetXOffset(const Value: GLfloat);
    procedure SetYOffset(const Value: GLfloat);
    procedure SetZoomX(const Value: GLfloat);
    procedure SetZoomY(const Value: GLfloat);
    procedure SetBrightness(const Value: Single);
    function GetOverlay(i: Integer): TGLOverlay;
   protected
    procedure glCreate;
    procedure glInit;
    procedure glLoadPicture(var aTexture: GLuint; const SizeX, SizeY: Integer; Pixels: Pointer;bpp:integer);
    procedure glDrawPicture;
    procedure Paint; override;
    procedure PanelResize(aSender: TObject);

    procedure hudmode(aoffsx,aoffsy,azoomx,azoomy:single);
    procedure popperspectivemode;
   public
    fTimings : array[  TBlitTimings] of integer;
    ftimers  : array[  TBlitTimings] of int64;
    fontsizecache : tfontsizecache;
    constructor Create(Owner: TComponent); override;
    constructor CreateFromPanel(aPanel: TPanel);
    destructor Destroy; override;
    function calcaspect:boolean;    
    procedure setoverrideaspect(b:boolean);
    function LoadPicture(const SizeX, SizeY: Integer; Pixels: Pointer;bpp:integer): Boolean;
    procedure overlayattach;
    procedure redrawafterchange;
    procedure setoverlays(Arrs: TOverLayData);
    procedure onceifattached;
    procedure flushoverlay;
    procedure attach(parant:TWincontrol);
    procedure detach;
    procedure clear;
    procedure clearoverlay;
    procedure setdetail(n:integer);
    procedure set2zoom(zx,zy:single);
    procedure set2offs(ox,oy:single);
    procedure setall (ox,oy,zx,zy:single;aspectoverride : boolean);
    procedure colorselection(col:Tcolor=clyellow);
    function  getdatainstring:ansistring;
    function zoomfitaspect(i: integer;var ox:single;var oy:single):single;
    function focuszoom(r: trect;var ox,oy:single):single;
    property Brightness: GLfloat read FBrightness write SetBrightness;
    property ZoomX: GLfloat read FZoomX write SetZoomX;
    property ZoomY: GLfloat read FZoomY write SetZoomY;
    property XOffset: GLfloat read FXOffset write SetXOffset;
    property YOffset: GLfloat read FYOffset write SetYOffset;
    property MaxTextureSize: GLint read FMaxTextureSize;
    property Overlay[i: Integer]: TGLOverlay read GetOverlay;
    property PictureWidth: Integer read FPictureWidth;
    property PictureHeight: Integer read FPictureHeight;
    property AspectOverride : boolean read foverrideaspect;
    property AspectX  : single read fAspectX;
    property AspectY  : single read fAspectY;
    property hud :TGLOverlay read fhud;
  end;

{ Debug time variables }

implementation

uses
  Math, SyncObjs;

var
  FirstTime: Boolean = True;
  CS: TCriticalSection;

{ TGLBlitter }

procedure TGLBlitter.SetXOffset(const Value: GLfloat);
begin
  FXOffset := Value;
  PanelResize(nil);
  Paint;
end;

procedure TGLBlitter.SetYOffset(const Value: GLfloat);
begin
  FYOffset := Value;
  PanelResize(nil);
  Paint;
end;

procedure TGLBlitter.SetZoomX(const Value: GLfloat);
begin
  FZoomX := Value;
  PanelResize(nil);
  Paint;
end;

procedure TGLBlitter.SetBrightness(const Value: Single);
const
  Tmp: array[0..3] of GLfloat = (-1, -1, -1, 1.0);
var
  i: Integer;
begin
  wglMakeCurrent(FDC, FRC);

  FBrightness := Value;

  for i := 0 to 2 do
    FLightAmbient[i] := Value;

  glEnable(GL_LIGHT1);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @Tmp[0]);
  glLightfv(GL_LIGHT1, GL_AMBIENT, @FLightAmbient[0]);
  glEnable(GL_LIGHTING);
end;

procedure TGLBlitter.HUDMODE(aoffsx,aoffsy,azoomx,azoomy:single);

begin
  glMatrixMode(GL_PROJECTION);					// Select Projection
	glPushMatrix();							// Push The Matrix
	glLoadIdentity();						// Reset The Matrix
  gluOrtho2D(aoffsx, 1.0 / (aZoomX) + (aOffsx), 1.0 / (aZoomY) + (aoffsy), aoffsy);
//	glOrtho( 0, x , y , 0, -1, 1 );				// Select Ortho Mode
	glMatrixMode(GL_MODELVIEW);					// Select Modelview Matrix
	glPushMatrix();							// Push The Matrix
	glLoadIdentity();
end;

procedure tglblitter.popperspectivemode;

begin
	// Set Up A Perspective View

	glMatrixMode( GL_PROJECTION );					// Select Projection
	glPopMatrix();							// Pop The Matrix
	glMatrixMode( GL_MODELVIEW );					// Select Modelview
	glPopMatrix();							// Pop The Matrix
end;

procedure TGLBlitter.PanelResize(aSender: TObject);
begin
  {$ifdef blitdebug}
  log.logline(srcdebug, inttostr(integer(self))+' '+floattostrf(fzoomx,fffixed,5,2)+' '+floattostrf(fzoomy,fffixed,5,2)+' '+floattostrf(fxoffset,fffixed,5,2)+' '+floattostrf(fyoffset,fffixed,5,2)+' '+floattostrf(faspectx,fffixed,5,2)+' '+floattostrf(faspecty,fffixed,5,2)+inttostr(fdc)+' '+inttostr(frc),catwarning);
  {$endif}
  if frc=0 then
    exit;
  
  starttimer(ftimers[blitrefresh2]);
  wglMakeCurrent(FDC, FRC);

  glViewport(0, 0, Width, Height);    // Set the viewport for the OpenGL window
  ftimings[blitrefresh2]:=endtimerus(ftimers[blitrefresh2]);

  starttimer(ftimers[blitrefresh3]);
  glMatrixMode(GL_PROJECTION);        // Change Matrix Mode to Projection
  glLoadIdentity();                   // Reset View
  ftimings[blitrefresh3]:=endtimerus(ftimers[blitrefresh3]);

  starttimer(ftimers[blitrefresh4]);
  calcaspect;             // aspect must be recalculated. needed for hudmode
  if foverrideaspect then
    gluOrtho2D(FXOffset, (1.0 / FZoomX) + FXOffset, (1.0 / FZoomY) + FYOffset, FYOffset)
  else
    gluOrtho2D(FXOffset, 1.0 / (FZoomX*faspectx) + (FXOffset), 1.0 / (FZoomY*faspecty) + (FYOffset), FYOffset);
  ftimings[blitrefresh4]:=endtimerus(ftimers[blitrefresh4]);

  glMatrixMode(GL_MODELVIEW);         // Return to the modelview matrix
end;

procedure TGLBlitter.glInit;
var b : bool;
begin
//$FFEEEF
  glClearColor( (clstdcolor and 255) / 256.0, ((clstdcolor shr 8) and 255) / 256.0, ((clstdcolor shr 16) and 255) / 256.0, 0.0);
  glEnable(GL_TEXTURE_RECTANGLE_ARB);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_DITHER);
  glDisable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glShadeModel(GL_SMOOTH); // was gl_smooth
  SetBrightness(FBrightness);

  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST); // was gl_nicenst

  glGetIntegerv(GL_MAX_TEXTURE_SIZE, @FMaxTextureSize);
end;

const xdiff = 0;
      ydiff = 0;

procedure TGLBlitter.glDrawPicture;
begin
  glClear(GL_COLOR_BUFFER_BIT);
//  glLoadIdentity;
  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, FTexture);

 if picloaded then
    begin

  glBegin(GL_QUADS);
   // bottom up
    glTexCoord2f(xdiff,PictureHeight+ydiff); glVertex2f(   0,    0);
    glTexCoord2f(FPictureWidth+xdiff, FPictureHeight+ydiff); glVertex2f( 1.0,    0);
    glTexCoord2f(FPictureWidth+xdiff, ydiff); glVertex2f( 1.0,  1.0);
    glTexCoord2f(xdiff, ydiff); glVertex2f(   0,  1.0);
   { //top down
       glTexCoord2f(0, 0); glVertex2f(   0,    0);
    glTexCoord2f(FPictureWidth, 0); glVertex2f( 1.0,    0);
    glTexCoord2f(FPictureWidth, FPictureHeight); glVertex2f( 1.0,  1.0);
    glTexCoord2f(0, FPictureHeight); glVertex2f(   0,  1.0);
    }
  glend;
  end;
end;

constructor TGLBlitter.Create(Owner: TComponent);
var i : integer;
begin
  inherited Create(Owner); // let's be owned by whoever owns the panel we're modelling after
  Parent := TWinControl(Owner);
  initdone:=false;
end;

destructor TGLBlitter.Destroy;
var
  i: Integer;
begin
  wglMakeCurrent(0, 0);
  for i := 1 to NrOverlayLayers do
    freeandnil(FOverlay[i]);
  freeandnil(FHud);
  wglDeleteContext(FRC);

  inherited;
end;

function TGLBlitter.LoadPicture(const SizeX, SizeY: Integer; Pixels: Pointer;bpp:integer): Boolean;
begin
  picloaded:=true;
  glLoadPicture(FTexture, SizeX, SizeY, Pixels,bpp);

//  Paint;

  Result := True;
end;

procedure TGLBlitter.Paint;
var
  i: Integer;
  x: Int64;
  b : bool;
  ox,oy,zx,zy : single;
begin
{  if (FPictureWidth = 0)
  or (FPictureHeight = 0) then
    Exit;
 }
  starttimer(x);
  {$ifdef debug}
  {$endif}
  starttimer(ftimers[blitrefresh3]);
  wglMakeCurrent(FDC, FRC);
  glDrawPicture;         // Draw the scene
  ftimings[blitrefresh3]:=endtimerus(ftimers[blitrefresh3]);

  starttimer(ftimers[blitrefresh4]);

//  doit:='';
  for i := 1 to NrOverlayLayers do
    begin
//      starttimer(x);
      FOverlay[i].DrawOverlay;                        // Draw the overlay
//      doit:=doit+inttostr(endtimerus(x))+' ';
    end;

 ftimings[blitrefresh4]:=endtimerus(ftimers[blitrefresh4]);

  starttimer(ftimers[blitrefresh1]);
  hudmode(0,0,aspectx,aspecty);
//  set2zoom(aspectx,aspecty);
//  set2offs(0,0);

//  PanelResize(nil);

  fhud.DrawOverlay;
  popperspectivemode;
  SwapBuffers(FDC);                  // Display the scene
  ftimings[blitrefresh1]:=endtimerus(ftimers[blitrefresh1]);
//  set2zoom(zx,zy);
//  set2offs(ox,oy);
  {$ifdef debug}
  {$endif}
end;

procedure TGLBlitter.glLoadPicture(var aTexture: GLuint; const SizeX,
  SizeY: Integer; Pixels: Pointer;bpp:integer);
var
  x: Int64;
  b: boolean;
begin
  b:=(FPictureWidth<>sizex) or (fpictureheight<>sizey);
  FPictureWidth := SizeX;
  FPictureHeight := SizeY;
  if b and not foverrideaspect then
    begin
      if calcaspect then
         redrawafterchange;
         sleep(10);
    end;

  b:=wglMakeCurrent(FDC, FRC);

  if textureallocated then
    glDeleteTextures(1,@atexture);
  glGenTextures(1, @aTexture); textureallocated:=true;

  glBindTexture(GL_TEXTURE_RECTANGLE_ARB, aTexture);
  {$ifdef debug}
  starttimer(x); // timercode, remove later
  {$endif}
  glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_rgb, SizeX, SizeY, 0,
               GL_BGR, GL_UNSIGNED_BYTE, Pixels);
  {$ifdef debug}
  ftimings[blittexture]:= endtimerus(x); // timercode, remove later
  {$endif}

  glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);	// was gl_linear=Linear Filtering. Dit is beter om pixels te kunnen zien
  glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);	// was gl_linear=Linear Filtering.
end;

function TGLBlitter.GetOverlay(i: Integer): TGLOverlay;
begin
  Result := FOverlay[i];
end;

procedure TGLBlitter.SetZoomY(const Value: GLfloat);
begin
  FZoomY := Value;
  PanelResize(nil);
  Paint;
end;

function TGLBlitter.zoomfitaspect(i: integer;var ox:single;var oy:single):single;
var ax,ay : single;
    ix,iy : integer;
begin
  faspectx:=PictureWidth/width;
  faspecty:=pictureheight/height;
  foverrideaspect:=false;

  ox:=0; oy:=0;

  ax:=1/(faspectx);
  ay:=1/(faspecty);
//  log.logline(srcdebug,'width:'+inttostr(width)+' height:'+inttostr(height),catwarning);
//  log.logline(srcdebug,'fpicturewidth:'+inttostr(fpicturewidth)+' fpictureheight:'+inttostr(fpictureheight),catwarning);
//  log.logline(srcdebug,'ax:'+FloatToStrf(ax,fffixed,2,5)+' ay:'+FloatToStrf(ay,fffixed,2,5),catwarning);
  if i=0  then
    begin
      if ax>ay then
        begin
          ox:=ay-ax;
          ax:=ay;
        end
      else
        oy:=ax-ay;
    end
  else
      begin
      if ax<ay then
        begin
          ox:=ay-ax;
          ax:=ay;
        end
      else
        oy:=ax-ay;
    end;
//  log.logline(srcdebug,'ax after max: '+FloatToStrf(ax,fffixed,2,5),catwarning);
  result:=ax;
end;

constructor TGLBlitter.CreateFromPanel(aPanel: TPanel);
begin
  inherited Create(aPanel.Owner);

  Parent := aPanel.Parent; // get panel's position etc. so we match it
  Align := aPanel.Align;
  Alignment := aPanel.Alignment;
  Anchors := aPanel.Anchors;
  Left := aPanel.Left;
  Top := aPanel.Top;
  Width := aPanel.Width; // seems this is ignored or something
  Height := aPanel.Height;
  Constraints.Assign(aPanel.Constraints);

  Enabled := True;
  Visible := True;

  aPanel.Visible := False; // don't draw the model panel
  faspectX:=1.0; faspectY:=1.0; foverrideaspect:=true;
  FZoomX := 1; // begin with 100%
  FZoomY := 1;
  
  glCreate;
  textureallocated:=false;
end;

procedure TGLBlitter.glCreate;
var
  pfd : TPIXELFORMATDESCRIPTOR;
  pf  : Integer;
  i   : Integer;
  b   : bool;
begin
  {$ifdef blitdebug}
  log.logline(srcdebug,'entering glcreate',catwarning);
  {$endif}
  if FirstTime then begin // used to do one time init stuff
    InitOpenGL;
    ReadExtensions;
  end;

  OnResize := PanelResize; // let's control our own resize, I have no idea which protected method is below this so...
  FBrightness := 5.0;

  FDC := GetDC(Handle);

  pfd.nSize := SizeOf(pfd);
  pfd.nVersion := 1;
  pfd.dwFlags := PFD_DRAW_TO_WINDOW or PFD_SUPPORT_OPENGL or PFD_DOUBLEBUFFER;
  pfd.iPixelType := PFD_TYPE_RGBA;
  pfd.cColorBits := 32;

  pf := ChoosePixelFormat(FDC, @pfd);
  b:=SetPixelFormat(FDC, pf, @pfd);

  FRC := wglCreateContext(FDC);
  b:=wglMakeCurrent(FDC, FRC);

  if FirstTime then begin
    ReadImplementationProperties;
    ReadExtensions;
  end;

  glInit;
  for i := 1 to NrOverlayLayers do
    begin
      if not assigned(foverlay[i]) then    
         FOverlay[i] := TGLOverlay.Create(Self);
      foverlay[i].Visible:=true;
      foverlay[i].Font.size:=fontsizecache[i];
    end;
  if not assigned(fhud) then
    FHUD := TGLOverlay.Create(Self);
  fHUD.Visible:=true;
  fhud.Font.size:=fontsizecache[0];
  {$ifdef blitdebug}
  log.logline(srcdebug,'glcreate before panelresize',catwarning);
  {$endif}
  PanelResize(nil);
  {$ifdef blitdebug}
  log.logline(srcdebug,'glcreate afterpanelresize',catwarning);
  {$endif}
  FirstTime := False;
end;

procedure TGLBlitter.setoverlays(Arrs: TOverLayData);
var i : integer;
begin
   for i:=low(arrs.layers) to high(arrs.layers) do
     Overlay[i].ovldata.assign(arrs.layers[i]);
end;

procedure TGLBlitter.onceifattached;
begin
  if not initdone then
    glCreate;
  initdone:=true;
end;

procedure TGLBlitter.flushoverlay;
var i : integer;
begin
 for i:=low(foverlay) to high(foverlay) do
   Overlay[i].Flush;
 fhud.Flush;
//    paint;
end;

function TGLBlitter.focuszoom(r: trect;var ox,oy:single):single;
var w,h: integer;
    r1,r2:single;
    nx,ny : single;
    ax,ay : single;
    pxw,pyh : single;
begin
 nx:=(r.Left)/(PictureWidth);    // offset indien linkerbovenhoek rect
 ny:=(r.Top)/(Pictureheight);    // = linkerbovenhoek viewport
 ax :=width/PictureWidth;        // zet local aspect mode.
 ay:= height/pictureheight;
 foverrideaspect:=false;

 w:=r.Right-r.left+1;            // breedte/hoogte rect
 h:=r.Bottom-r.Top+1;
 r1:=width/w;  r2:=height/h;     // bereken nodige zoom x/y
 if r1>r2 then r1:=r2;           // grootste waarde in r1 (zoom is 1/)

 r1:=trunc(r1);                  // round zoom up.
 pxw:=ax*picturewidth/r1;        // aantal pixels in breedte zichtbaar in viewport
 pyh:=ay*pictureheight/r1;       // aantal pixels in hoogte zichtbaar in viewport
 ox:=nx-0.5*(pxw-w)/picturewidth;   // verdeel verschil breedte rect <-> viewport over beide kanten
 oy:=ny-0.5*(pyh-h)/pictureheight;
 result:=r1;
end;

procedure TGLBlitter.attach(parant: TWincontrol);
var i : Integer;
begin
 glcreate;
 for i:=low(foverlay) to high(foverlay) do
   foverlay[i].attach;
 fhud.attach;
 textureallocated:=false;
end;

procedure TGLBlitter.detach;
var i : Integer;
begin
  for i:=low(foverlay) to high(foverlay) do
   foverlay[i].detach;
  fhud.detach;
  wglDeleteContext(frc); frc:=0;
  picloaded :=false;
end;

function TGLBlitter.calcaspect:boolean;
var nax,nay : single;
begin
  nax:=PictureWidth/width;
  nay:=pictureheight/height;
  result:=(abs(nax-faspectx) + abs(nay-faspecty))>0.001;
  faspectx:=nax;
  faspecty:=nay;  
end;

procedure TGLBlitter.setoverrideaspect(b: boolean);
begin
  if foverrideaspect<>b then
    begin
     foverrideaspect:=b;
     calcaspect;
//     PanelResize(nil);
    end;
end;

procedure TGLBlitter.set2zoom(zx, zy: single);
begin
 FZoomX:=zx;
 fzoomy:=zy;
end;

procedure TGLBlitter.setall(ox, oy, zx, zy: single; aspectoverride: boolean);
begin
 foverrideaspect:=aspectoverride;
 FZoomX:=zx;
 fzoomy:=zy;
 FXOffset:=ox;
 FYOffset:=oy;
end;

procedure TGLBlitter.redrawafterchange;
begin
  PanelResize(nil);
  Paint;
end;

procedure TGLBlitter.set2offs(ox, oy: single);
begin
  {$ifndef nobounds}
 if ox<0 then ox:=0;
 if ox>1 then ox:=1;
 if oy<0 then oy:=0;
 if oy>1 then oy:=1;
 {$endif}
 FXOffset:=ox;
 FYOffset:=oy; 
end;

procedure TGLBlitter.setdetail(n: integer);
var i : integer;
begin
 for i:=low(foverlay) to high(foverlay) do
     Overlay[i].Visible:=i<n;
end;

procedure TGLBlitter.clear;
begin
 picloaded:=false;
 glClear(GL_COLOR_BUFFER_BIT);
 clearoverlay;
 flushoverlay;
 paint;
end;

procedure TGLBlitter.clearoverlay;
var i : integer;
begin
 for i:=1 to high(foverlay) do
   if assigned(foverlay[i]) then
     foverlay[i].clear;
 if assigned(fhud) then
    fhud.clear;
end;

procedure TGLBlitter.overlayattach;
var i : integer;
begin
 for i:=low(foverlay) to high(foverlay) do
   begin
     foverlay[i].ffont.size:=fontsizecache[i];
     foverlay[i].attach;
   end;
 fhud.ffont.size:=fontsizecache[0];
 fhud.attach;
end;

function TGLBlitter.getdatainstring: ansistring;
begin
  result:='ZoomX:'+FloatToStrF(ZoomX,fffixed,5,3)+' ZoomY:'+FloatToStrF(ZoomY,fffixed,5,3)+' Offsx:'+FloatToStrF(fxoffset,fffixed,5,3)+'Offsy:'+FloatToStrF(fyoffset,fffixed,5,3);

end;

procedure TGLBlitter.colorselection(col: Tcolor);
var i : integer;
begin
  for i:=low(foverlay) to high(foverlay) do
   foverlay[i].colorselection(col);
 fhud.colorselection(col);
end;

{ TGLOverlay }

procedure TGLOverlay.BuildFont;
begin
  wglMakeCurrent(FOwner.FDC, FOwner.FRC);

  FFontBase := glGenLists(96);                            // 96 chars
  SelectObject(FOwner.FDC, FFont.Handle);                                // choose font to DC
  wglUseFontBitmaps(FOwner.FDC, 32, 96, FFontBase);                     // create 96 chars, start with 32 ASCII
end;

procedure TGLOverlay.KillFont;
begin
  glDeleteLists(FFontBase, 96); // free the font
end;

procedure TGLOverlay.SetFont(aFont: TFont);
begin
  FFont.Assign(aFont);

  wglMakeCurrent(FOwner.FDC, FOwner.FRC); // required, so we kill the proper font

  KillFont;
  BuildFont;

  FOwner.Paint;
end;

procedure TGLOverlay.SetVisible(const aValue: Boolean);
begin
  if FVisible = aValue then Exit;

  FVisible := aValue;

  FOwner.Paint;
end;

procedure TGLOverlay.TextOut(const i: Integer);
begin
  if not assigned(ovldata) Then exit;
  glEnable(GL_BLEND);
  glColor4ub(ovldata.FTextColors[i][1], ovldata.FTextColors[i][2], ovldata.FTextColors[i][3], ovldata.FTextColors[i][4]);
  glRasterPos2f(ovldata.FTextVertices[i][1], ovldata.FTextVertices[i][2]);
  glPrint(ovldata.FTexts[i]);
  glDisable(GL_BLEND);
end;

{ TGLOverlayVBO }

constructor TGLOverlay.Create(aOwner: TGLBlitter);
begin
  {inherited;}

  ovldata:=TOverlayLayerData.create;

  FOwner := aOwner;

  FFont := TFont.Create;
  FVisible := True;

  FLineWidth := 5;

  {end inherited}
  attach;

end;

destructor TGLOverlay.Destroy;
begin
  if fattached then
    detach;

//  KillFont; in detach
  {$ifdef dodeallocate}
  freeandnil(FFont);
  {$endif}
  ovldata.Free;
  inherited;
end;

procedure TGLOverlay.DrawOverlay;
var
  i: Integer;
  n : int64;
begin
  if not assigned(ovldata) then exit;
  if (not FVisible) then
    Exit;

  glDisable(GL_TEXTURE_RECTANGLE_ARB);
  glDisable(GL_LIGHTING);
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_COLOR_ARRAY);
  starttimer(n);
  if ovldata.FRectCount > 0 then begin
    { This assigns the data to specific "types and draw targets"
      so when GL reads our array/buffer it knows what to do with is }
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FRectVBO);
    glVertexPointer(2, GL_FLOAT, 0, nil); // 2 means 2 singles per vertex
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FRectCBO);
    glColorPointer(4, GL_UNSIGNED_BYTE, 0, nil); // 4 means 4 bytes per color of vertex

    glDrawArrays(GL_QUADS, 0, ovldata.FRectCount * 4); // this finally draws the rects,  *4 because each rect has 4 vertices
  end;
//  doit:=doit+'r: '+inttostr(endtimerus(n))  + '-';

  if ovldata.FTriangleCount > 0 then begin
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FTriangleVBO);
    glVertexPointer(2, GL_FLOAT, 0, nil); // 2 means 2 singles per vertex
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FTriangleCBO);
    glColorPointer(4, GL_UNSIGNED_BYTE, 0, nil); // 4 means 4 bytes per color of vertex

    glDrawArrays(GL_TRIANGLES, 0, ovldata.FTriangleCount * 3); // this finally draws the triangles,  *3 because each triangle has 3 vertices
  end;
  if ovldata.FStippleLineCount > 0 then begin
    glLineStipple(1,$00FF);    // dotted: $0101  dashed: $00FF   dash/dot/dash: $1C47
    glEnable(GL_LINE_STIPPLE);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FStippleLineVBO);
    glVertexPointer(2, GL_FLOAT, 0, nil); // 2 means 2 singles per vertex
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FStippleLineCBO);
    glColorPointer(4, GL_UNSIGNED_BYTE, 0, nil); // 4 means 4 bytes per color of vertex

    glDrawArrays(GL_LINES, 0, ovldata.FStippleLineCount * 2); // this finally draws the rects,  *4 because each line has 2 vertices
    glDisable(GL_LINE_STIPPLE);
  end;
  if ovldata.FThickLineCount > 0 then begin
    glLineWidth(FLineWidth); // set global line width for these babies
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FThickLineVBO);
    glVertexPointer(2, GL_FLOAT, 0, nil); // 2 means 2 singles per vertex
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FThickLineCBO);
    glColorPointer(4, GL_UNSIGNED_BYTE, 0, nil); // 4 means 4 bytes per color of vertex

    glDrawArrays(GL_LINES, 0, ovldata.FThickLineCount * 2); // this finally draws the rects,  *4 because each line has 2 vertices
    glLineWidth(1.0); // reset to 1px
  end;
  starttimer(n);
  if ovldata.FLineCount > 0 then begin
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FLineVBO);
    glVertexPointer(2, GL_FLOAT, 0, nil); // 2 means 2 singles per vertex
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FLineCBO);
    glColorPointer(4, GL_UNSIGNED_BYTE, 0, nil); // 4 means 4 bytes per color of vertex

    glDrawArrays(GL_LINES, 0, ovldata.FLineCount * 2); // this finally draws the rects,  *2 because each line has 2 vertices
  end;
//  doit:=doit+' t: '+'line: '+inttostr(ovldata.flinecount)+' = '+inttostr(endtimerus(n))+'-';
  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_COLOR_ARRAY);

  for i := 0 to ovldata.FTexts.Count - 1 do
    TextOut(i);

  glColor4ub(255, 255, 255, 0);
  glEnable(GL_TEXTURE_RECTANGLE_ARB);
  glEnable(GL_LIGHTING);
end;

procedure TGLOverlay.Flush;
  { Choose one of these and see if it makes impact.
    On my card it was 0 difference, but the docs say otherwise.
    However the friendly people at #opengl also say it doesn't really
    do any difference. I think todays implementations ignore it }
const
  METHOD = GL_STREAM_DRAW_ARB;
//  METHOD = GL_DYNAMIC_DRAW_ARB;
//  METHOD = GL_STATIC_DRAW_ARB;
var b : bool;
begin
  b:=wglMakeCurrent(FOwner.FDC, FOwner.FRC); // important

  if ovldata.FRectCount > 0 then begin
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FRectVBO);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, ovldata.FRectCount * SizeOf(TGLRect), @ovldata.FRects[0], METHOD);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FRectCBO);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, ovldata.FRectCount * SizeOf(TGLRectColor), @ovldata.FRectColors[0], METHOD);
  end;

  if ovldata.FLineCount > 0 then begin
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FLineVBO);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, ovldata.FLineCount * SizeOf(TGLLine), @ovldata.FLines[0], METHOD);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FlineCBO);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, ovldata.FLineCount * SizeOf(TGLLineColor), @ovldata.FLineColors[0], METHOD);
  end;

  if ovldata.FTriangleCount > 0 then begin
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FTriangleVBO);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, ovldata.FTriangleCount * SizeOf(TGLTriangle), @ovldata.FTriangles[0], METHOD);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FTriangleCBO);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, ovldata.FTriangleCount * SizeOf(TGLTriangleColor), @ovldata.FTriangleColors[0], METHOD);
  end;

  if ovldata.FThickLineCount > 0 then begin
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FThickLineVBO);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, ovldata.FThickLineCount * SizeOf(TGLLine), @ovldata.FThickLines[0], METHOD);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FThickLineCBO);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, ovldata.FThickLineCount * SizeOf(TGLLineColor), @ovldata.FThickLineColors[0], METHOD);
  end;
  
  if ovldata.FStippleLineCount > 0 then begin
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FStippleLineVBO);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, ovldata.FStippleLineCount * SizeOf(TGLLine), @ovldata.FStippleLines[0], METHOD);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, FStippleLineCBO);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, ovldata.FStippleLineCount * SizeOf(TGLLineColor), @ovldata.FStippleLineColors[0], METHOD);
  end;

//  FListStatus := glsFlushed;
end;

procedure TGLOverlay.glPrint(const aText: ansistring);
begin
  if aText = '' then
    Exit;

  glPushAttrib(GL_LIST_BIT);                              // save DL just in case
  glListBase(Max(0, FFontBase - 32));
  glCallLists(Length(aText), GL_UNSIGNED_BYTE, Pansichar(aText)); // draw the display list
  glPopAttrib;                                            // return to old state
end;

procedure TGLOverlay.attach;
begin
  BuildFont;
  glGenBuffersARB(1, @FRectVBO);
  glGenBuffersARB(1, @FRectCBO);

  glGenBuffersARB(1, @FLineVBO);
  glGenBuffersARB(1, @FLineCBO);

  glGenBuffersARB(1, @FThickLineVBO);
  glGenBuffersARB(1, @FThickLineCBO);

  glGenBuffersARB(1, @FStippleLineVBO);
  glGenBuffersARB(1, @FStippleLineCBO);

  glGenBuffersARB(1, @FTriangleVBO);
  glGenBuffersARB(1, @FTriangleCBO);
  fattached:=true;
end;

procedure TGLOverlay.detach;
begin
 clear;
 if fattached then
   begin
    KillFont;
    glDeleteBuffersARB(1, @FRectVBO);
    glDeleteBuffersARB(1, @FRectCBO);

    glDeleteBuffersARB(1, @FLineVBO);
    glDeleteBuffersARB(1, @FLineCBO);

    glDeleteBuffersARB(1, @FThickLineVBO);
    glDeleteBuffersARB(1, @FThickLineCBO);

    glDeleteBuffersARB(1, @FStippleLineVBO);
    glDeleteBuffersARB(1, @FStippleLineCBO);

    glDeleteBuffersARB(1, @FTriangleVBO);
    glDeleteBuffersARB(1, @FTriangleCBO);
    fattached:=false;
  end;
end;

procedure TGLOverlay.clear;
begin
  if assigned(ovldata) then
    ovldata.Clear;
end;

procedure TGLOverlay.assign(gl: TOverlaylayerdata);
begin
  ovldata.assign(gl);
end;

procedure TGLOverlay.colorselection(col: tcolor);
begin
 ovldata.colorselection(col);
end;

end.

